"""Provides the application configuration."""

from fastapi_factory_utilities.core.app import AppConfigAbstract


class AppConfig(AppConfigAbstract):
    """The application configuration.

    Values are set in application.yaml.
    """

    pass



"""Provides an abstract class for FastAPI application integration."""

from abc import ABC
from typing import Any

import starlette.types
from fastapi import APIRouter, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, ConfigDict, Field


class FastAPIConfigAbstract(ABC, BaseModel):
    """Partial configuration for FastAPI."""

    model_config = ConfigDict(strict=False)

    # Application metadata
    title: str
    description: str
    version: str

    # Host configuration
    host: str = Field(default="0.0.0.0")
    port: int = Field(default=8000)

    # Root configuration
    root_path: str = Field(default="")

    # Debug mode
    debug: bool = Field(default=False, strict=False)

    # Uvicorn configuration
    reload: bool = Field(default=False, strict=False)
    workers: int = Field(default=1, strict=False)


class FastAPIAbstract(ABC):
    """Application integration with FastAPI.

    TODO: Replace by a Factory pattern.
    """

    def __init__(
        self,
        config: FastAPIConfigAbstract,
        api_router: APIRouter | None = None,
        lifespan: starlette.types.StatelessLifespan[starlette.types.ASGIApp] | None = None,
    ) -> None:
        """Instanciate the FastAPI application.

        Args:
            config (FastAPIConfigAbstract): The FastAPI configuration.
            api_router (APIRouter, optional): The API router to include.
                Defaults to None.
            lifespan (AsyncGenerator[None, None], optional): The lifespan

        Returns:
            None

        """
        self._fastapi_app: FastAPI = FastAPI(
            title=config.title,
            description=config.description,
            version=config.version,
            root_path=config.root_path,
            debug=config.debug,
            lifespan=lifespan,
        )

        # TODO: Add CORS middleware Configuration
        self._fastapi_app.add_middleware(
            middleware_class=CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        if api_router is not None:
            self._fastapi_app.include_router(router=api_router)

    def get_asgi_app(self) -> FastAPI:
        """Get the ASGI application."""
        return self._fastapi_app

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        """Forward the call to the FastAPI app."""
        return await self._fastapi_app.__call__(scope=scope, receive=receive, send=send)



"""Provides the abstract class for the application."""

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import ClassVar, Self, cast

import starlette.types
from beanie import Document
from fastapi import FastAPI

from fastapi_factory_utilities.core.api import api
from fastapi_factory_utilities.core.utils.log import LogModeEnum, setup_log

from .config_abstract import AppConfigAbstract, AppConfigBuilder
from .fastapi_application_abstract import FastAPIAbstract
from .plugins_manager_abstract import (
    ApplicationPluginManagerAbstract,
    PluginsActivationList,
)


class BaseApplication(FastAPIAbstract, ApplicationPluginManagerAbstract):
    """Application abstract class."""

    PACKAGE_NAME: str = ""

    CONFIG_CLASS: ClassVar[type[AppConfigAbstract]] = AppConfigAbstract

    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]] = []

    def __init__(self, config: AppConfigAbstract, plugin_activation_list: PluginsActivationList | None = None) -> None:
        """Instantiate the application.

        Args:
            config (AppConfigAbstract): The application configuration.
            plugin_activation_list (PluginsActivationList | None, optional): The plugins activation list.

        Returns:
            None

        Raises:
            ValueError: If the package name is not set.
        """
        if self.PACKAGE_NAME == "":
            raise ValueError("The package name must be set in the concrete application class.")

        self._config: AppConfigAbstract = config
        FastAPIAbstract.__init__(
            self=cast(FastAPIAbstract, self),
            config=self._config,
            api_router=api,
            lifespan=cast(starlette.types.StatelessLifespan[starlette.types.ASGIApp], self.fastapi_lifespan),
        )
        ApplicationPluginManagerAbstract.__init__(
            self=cast(ApplicationPluginManagerAbstract, self), plugin_activation_list=plugin_activation_list
        )
        self._on_load()

    @classmethod
    def main(cls) -> None:
        """Main function.

        This must be the same for all applications.
        """
        from fastapi_factory_utilities.core.utils.uvicorn import (  # pylint: disable=import-outside-toplevel
            UvicornUtils,
        )

        setup_log(mode=LogModeEnum.CONSOLE)
        application: BaseApplication = cls.build()
        uvicorn_utils = UvicornUtils(app=application)

        try:
            uvicorn_utils.serve()
        except KeyboardInterrupt:
            pass

    @classmethod
    def build_config(cls) -> AppConfigAbstract:
        """Build the application configuration.

        Returns:
            AppConfigAbstract: The application configuration.
        """
        config_builder: AppConfigBuilder = AppConfigBuilder(
            package_name=cls.PACKAGE_NAME, config_class=cls.CONFIG_CLASS
        )
        return config_builder.build()

    @classmethod
    def build(
        cls, config: AppConfigAbstract | None = None, plugin_activation_list: PluginsActivationList | None = None
    ) -> Self:
        """Build the application.

        Args:
            config (AppConfigAbstract | None, optional): The application configuration. Defaults to None.
            plugin_activation_list (PluginsActivationList | None, optional): The plugins activation list.
            Defaults to None.
        """
        if config is None:
            config = cls.build_config()

        return cls(config=config, plugin_activation_list=plugin_activation_list)

    @asynccontextmanager
    async def fastapi_lifespan(self, fastapi_application: FastAPI) -> AsyncGenerator[None, None]:
        """Provide the lifespan context manager for FastAPI.

        Args:
            fastapi_application (FastAPI): The FastAPI application.

        Returns:
            AsyncGenerator[None]: The lifespan context manager.
        """
        del fastapi_application
        await self.plugins_on_startup()
        yield
        await self.plugins_on_shutdown()

    def get_config(self) -> AppConfigAbstract:
        """Get the application configuration."""
        return self._config



"""Provide the configuration for the app server."""

from pydantic import Field

from fastapi_factory_utilities.core.app.base.exceptions import (
    ApplicationConfigFactoryException,
)
from fastapi_factory_utilities.core.app.base.plugins_manager_abstract import (
    PluginsActivationList,
)
from fastapi_factory_utilities.core.utils.configs import (
    UnableToReadConfigFileError,
    ValueErrorConfigError,
    build_config_from_file_in_package,
)
from fastapi_factory_utilities.core.utils.log import LoggingConfig

from ..enums import EnvironmentEnum
from .fastapi_application_abstract import FastAPIConfigAbstract


class AppConfigAbstract(FastAPIConfigAbstract, PluginsActivationList):
    """Application configuration abstract class."""

    environment: EnvironmentEnum
    service_name: str
    service_namespace: str

    logging: list[LoggingConfig] = Field(default_factory=list, description="Logging configuration.")


class AppConfigBuilder:
    """Application configuration builder."""

    DEFAULT_FILENAME: str = "application.yaml"
    DEFAULT_YAML_BASE_KEY: str = "application"

    def __init__(
        self,
        package_name: str,
        config_class: type[AppConfigAbstract],
        filename: str = DEFAULT_FILENAME,
        yaml_base_key: str = DEFAULT_YAML_BASE_KEY,
    ) -> None:
        """Instantiate the builder.

        Args:
            package_name (str): The package name.
            config_class (Type[AppConfigAbstract]): The configuration class.
            filename (str, optional): The filename. Defaults to DEFAULT_FILENAME.
            yaml_base_key (str, optional): The YAML base key. Defaults to DEFAULT_YAML_BASE_KEY.
        """
        self.package_name: str = package_name
        self.config_class: type[AppConfigAbstract] = config_class
        self.filename: str = filename
        self.yaml_base_key: str = yaml_base_key

    def build(self) -> AppConfigAbstract:
        """Build the configuration.

        Returns:
            AppConfigAbstract: The configuration.
        """
        try:
            config: AppConfigAbstract = build_config_from_file_in_package(
                package_name=self.package_name,
                config_class=self.config_class,
                filename=self.filename,
                yaml_base_key=self.yaml_base_key,
            )
        except UnableToReadConfigFileError as exception:
            raise ApplicationConfigFactoryException("Unable to read the application configuration file.") from exception
        except ValueErrorConfigError as exception:
            raise ApplicationConfigFactoryException(
                "Unable to create the application configuration model."
            ) from exception

        return config



"""Provides the concrete application class."""

from typing import ClassVar

from beanie import Document

from fastapi_factory_utilities.core.app import BaseApplication
from fastapi_factory_utilities.core.app.base.plugins_manager_abstract import (
    PluginsActivationList,
)
from fastapi_factory_utilities.example.models.books.document import BookDocument

from .config import AppConfig


class App(BaseApplication):
    """Concrete application class."""

    PACKAGE_NAME: str = "fastapi_factory_utilities.example"

    CONFIG_CLASS = AppConfig

    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]] = [BookDocument]

    def __init__(self, config: AppConfig, plugin_activation_list: PluginsActivationList | None = None) -> None:
        """Instantiate the application with the configuration and the API router.

        Args:
            config (AppConfig): The application configuration.
            plugin_activation_list (PluginsActivationList | None, optional): The plugins activation list.
        """
        super().__init__(config=config, plugin_activation_list=plugin_activation_list)

        # Prevent circular imports
        from ..api import api_router  # pylint: disable=import-outside-toplevel

        self.get_asgi_app().include_router(router=api_router)



"""Provides the exceptions for the application factory."""


class BaseApplicationException(BaseException):
    """Base application exception."""

    pass


class ApplicationFactoryException(BaseApplicationException):
    """Application factory exception."""

    pass


class ApplicationConfigFactoryException(BaseApplicationException):
    """Application configuration factory exception."""

    pass


class ApplicationPluginManagerException(BaseApplicationException):
    """Application plugin manager exception."""

    pass



"""Provides utilities for the application."""

import uvicorn
import uvicorn.server

from fastapi_factory_utilities.core.protocols import BaseApplicationProtocol
from fastapi_factory_utilities.core.utils.log import clean_uvicorn_logger


class UvicornUtils:
    """Provides utilities for Uvicorn."""

    def __init__(self, app: BaseApplicationProtocol) -> None:
        """Instantiate the factory.

        Args:
            app (BaseApplication): The application.
            config (AppConfigAbstract): The application configuration.

        Returns:
            None
        """
        self._app: BaseApplicationProtocol = app

    def build_uvicorn_config(self) -> uvicorn.Config:
        """Build the Uvicorn configuration.

        Returns:
            uvicorn.Config: The Uvicorn configuration.
        """
        config = uvicorn.Config(
            app=self._app.get_asgi_app(),
            host=self._app.get_config().host,
            port=self._app.get_config().port,
            reload=self._app.get_config().reload,
            workers=self._app.get_config().workers,
        )
        clean_uvicorn_logger()
        return config

    def serve(self) -> None:
        """Serve the application."""
        config: uvicorn.Config = self.build_uvicorn_config()
        server: uvicorn.Server = uvicorn.Server(config=config)
        server.run()



"""API v1 sys health module.

Provide the Get health endpoint
"""

from enum import StrEnum
from http import HTTPStatus

from fastapi import APIRouter, Response
from pydantic import BaseModel

api_v1_sys_health = APIRouter(prefix="/health")


class HealthStatusEnum(StrEnum):
    """Health status enum."""

    HEALTHY = "healthy"
    UNHEALTHY = "unhealthy"


class HealthResponseModel(BaseModel):
    """Health response schema."""

    status: HealthStatusEnum


@api_v1_sys_health.get(
    path="",
    tags=["sys"],
    response_model=HealthResponseModel,
    responses={
        HTTPStatus.OK.value: {
            "model": HealthResponseModel,
            "description": "Health status.",
        },
        HTTPStatus.INTERNAL_SERVER_ERROR.value: {
            "model": HealthResponseModel,
            "description": "Internal server error.",
        },
    },
)
def get_api_v1_sys_health(response: Response) -> HealthResponseModel:
    """Get the health of the system.

    Args:
        response (Response): The response object.

    Returns:
        HealthResponse: The health status.
    """
    response.status_code = HTTPStatus.OK
    return HealthResponseModel(status=HealthStatusEnum.HEALTHY)



"""Provides the core application module for the Python Factory."""

from .base import (
    AppConfigAbstract,
    ApplicationConfigFactoryException,
    ApplicationFactoryException,
    BaseApplication,
    BaseApplicationException,
)
from .enums import EnvironmentEnum

__all__: list[str] = [
    "BaseApplication",
    "AppConfigAbstract",
    "EnvironmentEnum",
    "ApplicationConfigFactoryException",
    "ApplicationFactoryException",
    "BaseApplicationException",
]



"""Oriented Data Model (ODM) plugin package."""

from logging import INFO, Logger, getLogger
from typing import Any

from beanie import init_beanie  # pyright: ignore[reportUnknownVariableType]
from motor.motor_asyncio import AsyncIOMotorClient
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.protocols import BaseApplicationProtocol

from .builder import ODMBuilder

_logger: BoundLogger = get_logger()


def pre_conditions_check(application: BaseApplicationProtocol) -> bool:
    """Check the pre-conditions for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        bool: True if the pre-conditions are met, False otherwise.
    """
    del application
    return True


def on_load(
    application: BaseApplicationProtocol,
) -> None:
    """Actions to perform on load for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.
    """
    del application
    # Configure the pymongo logger to INFO level
    pymongo_logger: Logger = getLogger("pymongo")
    pymongo_logger.setLevel(INFO)
    _logger.debug("ODM plugin loaded.")


async def on_startup(
    application: BaseApplicationProtocol,
) -> None:
    """Actions to perform on startup for the ODM plugin.

    Args:
        application (BaseApplicationProtocol): The application.
        odm_config (ODMConfig): The ODM configuration.

    Returns:
        None
    """
    try:
        odm_factory: ODMBuilder = ODMBuilder(application=application).build_all()
    except Exception as exception:  # pylint: disable=broad-except
        _logger.error(f"ODM plugin failed to start. {exception}")
        return

    if odm_factory.odm_database is None or odm_factory.odm_client is None:
        _logger.error(
            f"ODM plugin failed to start. Database: {odm_factory.odm_database} - " f"Client: {odm_factory.odm_client}"
        )
        return
    # TODO: Find a way to add type to the state
    application.get_asgi_app().state.odm_client = odm_factory.odm_client
    application.get_asgi_app().state.odm_database = odm_factory.odm_database

    # TODO: Find a better way to initialize beanie with the document models of the concrete application
    # through an hook in the application ?
    await init_beanie(
        database=odm_factory.odm_database,
        document_models=application.ODM_DOCUMENT_MODELS,
    )

    _logger.info(
        f"ODM plugin started. Database: {odm_factory.odm_database.name} - "
        f"Client: {odm_factory.odm_client.address} - "
        f"Document models: {application.ODM_DOCUMENT_MODELS}"
    )


async def on_shutdown(application: BaseApplicationProtocol) -> None:
    """Actions to perform on shutdown for the ODM plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        None
    """
    client: AsyncIOMotorClient[Any] = application.get_asgi_app().state.odm_client
    client.close()
    _logger.debug("ODM plugin shutdown.")



"""Provides the module for the ODM plugin."""

import asyncio
import time
from typing import Any, Self

from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from structlog.stdlib import get_logger

from fastapi_factory_utilities.core.protocols import BaseApplicationProtocol
from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import (
    UnableToReadYamlFileError,
    YamlFileReader,
)

from .configs import ODMConfig
from .exceptions import ODMPluginConfigError

_logger = get_logger()


class ODMBuilder:
    """Factory to create the resources for the ODM plugin.

    The factory is responsible for creating the resources for the ODM plugin.
    - The ODM configuration.
    - The ODM client.
    - The ODM database.

    ```python
    # Example of using the ODMFactory
    odm_factory: ODMFactory = ODMFactory(application=application)
    odm_factory.build_all()
    # Access the ODM database created
    database: AsyncIOMotorDatabase[Any] = odm_factory.database
    ```

    """

    def __init__(
        self,
        application: BaseApplicationProtocol,
        odm_config: ODMConfig | None = None,
        odm_client: AsyncIOMotorClient[Any] | None = None,
        odm_database: AsyncIOMotorDatabase[Any] | None = None,
    ) -> None:
        """Initialize the ODMFactory.

        Args:
            application (BaseApplicationProtocol): The application.
            odm_config (ODMConfig): The ODM configuration for injection. (Default is None)
            odm_client (AsyncIOMotorClient): The ODM client for injection. (Default is None)
            odm_database (AsyncIOMotorDatabase): The ODM database for injection. (Default is None)

        """
        self._application: BaseApplicationProtocol = application
        self._config: ODMConfig | None = odm_config
        self._odm_client: AsyncIOMotorClient[Any] | None = odm_client
        self._odm_database: AsyncIOMotorDatabase[Any] | None = odm_database

    @property
    def config(self) -> ODMConfig | None:
        """Provide the ODM configuration object.

        Returns:
            ODMConfig: The ODM configuration object.
        """
        return self._config

    @property
    def odm_client(self) -> AsyncIOMotorClient[Any] | None:
        """Provide the ODM client.

        Returns:
            AsyncIOMotorClient | None: The ODM client.
        """
        return self._odm_client

    @property
    def odm_database(self) -> AsyncIOMotorDatabase[Any] | None:
        """Provide the ODM database.

        Returns:
            AsyncIOMotorDatabase | None: The ODM database.
        """
        return self._odm_database

    def build_odm_config(
        self,
    ) -> Self:
        """Build the ODM configuration object.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the package name is not set or the configuration file is not found.
        """
        if self._config is not None:
            return self

        if self._application.PACKAGE_NAME == "":
            raise ODMPluginConfigError("The package name must be set in the concrete application class.")
        # Read the application configuration file
        try:
            yaml_file_content: dict[str, Any] = YamlFileReader(
                file_path=get_path_file_in_package(
                    filename="application.yaml",
                    package=self._application.PACKAGE_NAME,
                ),
                yaml_base_key="odm",
                use_environment_injection=True,
            ).read()
        except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
            raise ODMPluginConfigError("Unable to read the application configuration file.") from exception

        # Create the application configuration model
        try:
            self._config = ODMConfig(**yaml_file_content)
        except ValueError as exception:
            raise ODMPluginConfigError("Unable to create the application configuration model.") from exception
        return self

    @classmethod
    def _wait_client_to_be_ready(cls, client: AsyncIOMotorClient[Any], timeout_ms: int) -> None:
        """Wait for the ODM client to be ready.

        Args:
            client (AsyncIOMotorClient): The ODM client.
            timeout_ms (int): The timeout in milliseconds.

        Raises:
            ODMPluginConfigError: If the ODM client is not ready.
        """
        start_timer = time.monotonic()

        async def is_connected(client: AsyncIOMotorClient[Any]) -> bool:
            """Check if the client is connected."""
            try:
                await client.admin.command(
                    command="ping",
                )  # pyright: ignore
                return True
            except Exception:  # pylint: disable=broad-except
                _logger.debug("ODM client is not ready.")
                return False

        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

        while not loop.run_in_executor(None, asyncio.Task(is_connected(client))) and (  # type: ignore
            (time.monotonic() - start_timer) < timeout_ms
        ):
            if time.monotonic() - start_timer > timeout_ms:
                raise ODMPluginConfigError("ODM client is not ready.")
            time.sleep(0.01)

        if not loop.run_in_executor(None, asyncio.Task(is_connected(client))):  # type: ignore
            raise ODMPluginConfigError("ODM client is not ready.")

    def build_client(
        self,
    ) -> Self:
        """Build the ODM client.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        if self._odm_client is not None:
            return self

        if self._config is None:
            raise ODMPluginConfigError(
                "ODM configuration is not set. Provide the ODM configuration using "
                "build_odm_config method or through parameter."
            )

        self._odm_client = AsyncIOMotorClient(
            host=self._config.uri,
            connect=True,
            connectTimeoutMS=self._config.connection_timeout_ms,
            serverSelectionTimeoutMS=self._config.connection_timeout_ms,
        )

        self._wait_client_to_be_ready(client=self._odm_client, timeout_ms=self._config.connection_timeout_ms)

        return self

    def build_database(
        self,
    ) -> Self:
        """Build the ODM database.

        The ODM client and ODM configuration are recommended to be provided through call to the build_client and
        build_odm_config methods.

        Returns:
            Any: The ODM database.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        if self._odm_database is not None:
            return self

        if self._config is None:
            raise ODMPluginConfigError(
                "ODM configuration is not set. Provide the ODM configuration using "
                "build_odm_config method or through parameter."
            )

        database_name: str = self._config.database

        if self._odm_client is None:
            raise ODMPluginConfigError(
                "ODM client is not set. Provide the ODM client using " "build_client method or through parameter."
            )

        self._odm_database = self._odm_client.get_database(name=database_name)

        return self

    def build_all(self) -> Self:
        """Build all the resources for the ODM plugin.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        self.build_odm_config()
        self.build_client()
        self.build_database()

        return self



"""Package for the base application, abstract config classes and related exceptions."""

from .application import BaseApplication
from .config_abstract import AppConfigAbstract
from .exceptions import (
    ApplicationConfigFactoryException,
    ApplicationFactoryException,
    BaseApplicationException,
)

__all__: list[str] = [
    "BaseApplication",
    "AppConfigAbstract",
    "ApplicationConfigFactoryException",
    "ApplicationFactoryException",
    "BaseApplicationException",
]



"""Plugins manager abstract module."""

from abc import ABC
from importlib import import_module
from types import ModuleType
from typing import cast

from pydantic import BaseModel, ConfigDict, Field

from fastapi_factory_utilities.core.plugins import PluginsEnum
from fastapi_factory_utilities.core.protocols import (
    BaseApplicationProtocol,
    PluginProtocol,
)
from fastapi_factory_utilities.core.utils.configs import (
    UnableToReadConfigFileError,
    ValueErrorConfigError,
    build_config_from_file_in_package,
)

from .exceptions import ApplicationPluginManagerException


class PluginsActivationList(BaseModel):
    """Model for the plugins activation list."""

    model_config = ConfigDict(extra="forbid")

    activate: list[PluginsEnum] = Field(default=list())


class ApplicationPluginManagerAbstract(ABC):
    """Abstract class for the application plugin manager.

    Responsibilities:
    - Retrieve the plugins for the application.
    - Check the pre-conditions for the plugins.
    - Perform actions on startup for the plugins.
    - Perform actions on shutdown for the plugins.
    """

    PACKAGE_NAME: str = ""

    PLUGIN_PACKAGE_NAME: str = "fastapi_factory_utilities.core.plugins"

    def __init__(self, plugin_activation_list: PluginsActivationList | None = None) -> None:
        """Instanciate the application plugin manager."""
        if self.PACKAGE_NAME == "":
            raise ValueError("The package name must be set in the concrete plugin manager class.")

        self._plugins: list[PluginProtocol] = []
        self._plugins_activation_list: PluginsActivationList
        if plugin_activation_list is not None:
            self._plugins_activation_list = plugin_activation_list
        else:
            self._plugins_activation_list = self._build_plugins_activation_list()

        self._check_pre_conditions()

    def _check_pre_conditions(self) -> None:
        """Check the pre-conditions for the plugins.

        Raises:
            ApplicationPluginManagerException: If a plugin is not
            activated.

        """
        for plugin in self._plugins_activation_list.activate:
            try:
                plugin_module: ModuleType = import_module(name=f"{self.PLUGIN_PACKAGE_NAME}.{plugin.value}")
            except ImportError as exception:
                raise ApplicationPluginManagerException(f"Unable to import the plugin {plugin.value}") from exception

            if not isinstance(plugin_module, PluginProtocol):
                raise ApplicationPluginManagerException(
                    f"The plugin {plugin.value} does not implement the PluginProtocol"
                )

            if not plugin_module.pre_conditions_check(application=cast(BaseApplicationProtocol, self)):
                raise ApplicationPluginManagerException(f"The plugin {plugin.value} does not meet the pre-conditions")

            self._plugins.append(plugin_module)

    def _on_load(self) -> None:
        """Actions to perform on load for the plugins."""
        for plugin in self._plugins:
            plugin.on_load(application=cast(BaseApplicationProtocol, self))

    def _build_plugins_activation_list(self) -> PluginsActivationList:
        """Build the plugins activation list.

        Returns:
            PluginsActivationList: The plugins activation list.

        Raises:
            ApplicationPluginManagerException: If there is an error
            reading the configuration file.
            ApplicationPluginManagerException: If there is an error
            creating the configuration model.

        """
        try:
            config: PluginsActivationList = build_config_from_file_in_package(
                package_name=self.PACKAGE_NAME,
                filename="application.yaml",
                config_class=PluginsActivationList,
                yaml_base_key="plugins",
            )
        except UnableToReadConfigFileError as exception:
            raise ApplicationPluginManagerException("Unable to read the application configuration file") from exception
        except ValueErrorConfigError as exception:
            raise ApplicationPluginManagerException(
                "Unable to create the application configuration model"
            ) from exception

        return config

    async def plugins_on_startup(self) -> None:
        """Actions to perform on startup for the plugins."""
        for plugin in self._plugins:
            try:
                await plugin.on_startup(application=cast(BaseApplicationProtocol, self))
            except Exception as exception:
                raise ApplicationPluginManagerException(
                    f"Error during the startup of the plugin {plugin.__class__.__name__}"
                ) from exception

    async def plugins_on_shutdown(self) -> None:
        """Actions to perform on shutdown for the plugins."""
        for plugin in self._plugins:
            try:
                await plugin.on_shutdown(application=cast(BaseApplicationProtocol, self))
            except Exception as exception:
                raise ApplicationPluginManagerException(
                    f"Error during the shutdown of the plugin {plugin.__class__.__name__}"
                ) from exception



"""API v1 sys readiness module.

Provide the Get readiness endpoint
"""

from enum import StrEnum
from http import HTTPStatus

from fastapi import APIRouter, Response
from pydantic import BaseModel

api_v1_sys_readiness = APIRouter(prefix="/readiness")


class ReadinessStatusEnum(StrEnum):
    """Readiness status enum."""

    READY = "ready"
    NOT_READY = "not_ready"


class ReadinessResponseModel(BaseModel):
    """Readiness response schema."""

    status: ReadinessStatusEnum


@api_v1_sys_readiness.get(
    path="",
    tags=["sys"],
    response_model=ReadinessResponseModel,
    responses={
        HTTPStatus.OK.value: {
            "model": ReadinessResponseModel,
            "description": "Readiness status.",
        },
        HTTPStatus.INTERNAL_SERVER_ERROR.value: {
            "model": ReadinessResponseModel,
            "description": "Internal server error.",
        },
    },
)
def get_api_v1_sys_readiness(response: Response) -> ReadinessResponseModel:
    """Get the readiness of the system.

    Args:
        response (Response): The response object.

    Returns:
        ReadinessResponse: The readiness status.
    """
    response.status_code = HTTPStatus.OK
    return ReadinessResponseModel(status=ReadinessStatusEnum.READY)



"""Package for plugins."""

from enum import StrEnum, auto


class PluginsEnum(StrEnum):
    """Enumeration for the plugins."""

    OPENTELEMETRY_PLUGIN = auto()
    ODM_PLUGIN = auto()


__all__: list[str] = [
    "PluginsEnum",
]
